/*
 * brutebrand: This program brutally changes the OS ABI identification byte in
 * an ELF binary. syntax: ./brutebrand -t type file
 *
 * Copyright (C) 2004 shadowinteger
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *
 * $Id: brutebrand.c,v 1.3 2004/02/21 12:21:55 shadow Exp $
 */

/* If you have /usr/include/elf.h and rather want to use that instead,
 * uncomment the line below or compile with -DHAVEELFH
 */
//#define HAVEELFH


#include <stdio.h>
#include <unistd.h>
#include <libgen.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

extern char *optarg;
extern int optind, opterr, optopt;

extern int errno;

#ifdef HAVEELFH
    #include <elf.h>
#else
    /* these were grabbed from /usr/include/elf.h */
    #define EI_OSABI        7               /* OS ABI identification */
    #define ELFOSABI_NONE           0       /* UNIX System V ABI */
    #define ELFOSABI_SYSV           0       /* Alias.  */
    #define ELFOSABI_HPUX           1       /* HP-UX */
    #define ELFOSABI_NETBSD         2       /* NetBSD.  */
    #define ELFOSABI_LINUX          3       /* Linux.  */
    #define ELFOSABI_SOLARIS        6       /* Sun Solaris.  */
    #define ELFOSABI_AIX            7       /* IBM AIX.  */
    #define ELFOSABI_IRIX           8       /* SGI Irix.  */
    #define ELFOSABI_FREEBSD        9       /* FreeBSD.  */
    #define ELFOSABI_TRU64          10      /* Compaq TRU64 UNIX.  */
    #define ELFOSABI_MODESTO        11      /* Novell Modesto.  */
    #define ELFOSABI_OPENBSD        12      /* OpenBSD.  */
    #define ELFOSABI_ARM            97      /* ARM */
    #define ELFOSABI_STANDALONE     255     /* Standalone (embedded) application */
#endif

void usage(char *p) {
    printf("usage: %s [-h] -t type elf_file...\n"
           "type can be one of these (by name):\n"
           "    none, sysv, hpux, netbsd, linux, solaris, aix, irix,\n"
           "    freebsd, tru64, modesto, openbsd, arm, standalone.\n",
           p);
    return;
}

int main(int argc, char **argv) {
    int i;
    unsigned char type, got_type = 0;
    unsigned int elf;
    char *p;

    p = strdup(argv[0]);
    p = basename(p);

    while ((i = getopt(argc, argv, "ht:")) != EOF) {
        switch(i) {
            case 'h':
                usage(p);
                return 1;
            case 't':
                if (!strcasecmp("none", optarg))
                    type = ELFOSABI_NONE;
                else if (!strcasecmp("sysv", optarg))
                    type = ELFOSABI_SYSV;
                else if (!strcasecmp("hpux", optarg))
                    type = ELFOSABI_HPUX;
                else if (!strcasecmp("netbsd", optarg))
                    type = ELFOSABI_NETBSD;
                else if (!strcasecmp("linux", optarg))
                    type = ELFOSABI_LINUX;
                else if (!strcasecmp("solaris", optarg))
                    type = ELFOSABI_SOLARIS;
                else if (!strcasecmp("aix", optarg))
                    type = ELFOSABI_AIX;
                else if (!strcasecmp("irix", optarg))
                    type = ELFOSABI_IRIX;
                else if (!strcasecmp("freebsd", optarg))
                    type = ELFOSABI_FREEBSD;
                else if (!strcasecmp("tru64", optarg))
                    type = ELFOSABI_TRU64;
                else if (!strcasecmp("modesto", optarg))
                    type = ELFOSABI_MODESTO;
                else if (!strcasecmp("openbsd", optarg))
                    type = ELFOSABI_OPENBSD;
                else if (!strcasecmp("arm", optarg))
                    type = ELFOSABI_ARM;
                else if (!strcasecmp("standalone", optarg))
                    type = ELFOSABI_STANDALONE;
                else {
                    fprintf(stderr, "%s: unknown type: \"%s\"\n", p, optarg);
                    return 1;
                }
                got_type = 1;
                break;
            default:
                usage(p);
                return 1;
        }
     }

    if (!got_type) {
        usage(p);
        return 1;
    }

    if (optind < argc) {
        while (optind < argc) {
            /* brutebrand loop */
            if ((i = open(argv[optind], O_RDWR)) < 0) {
                fprintf(stderr, "%s: %s: %s\n", p, argv[optind], strerror(errno));
                return 1;
            }

            /* read elf magic number */
            if (read(i, &elf, 4) < 0) { /* I used sizeof(elf) here before */
                fprintf(stderr, "%s: %s: %s\n", p, argv[optind], strerror(errno));
                return 1;
            }
            /* check if it's an elf, if not, don't modify */
            if (elf != 0x464c457f) {    /* yes, that's ELF cruft */
                fprintf(stderr, "%s: won't modify %s, not and ELF binary\n", p, argv[optind]);
            } else {
                /* seek to EI_OSABI (7) */
                if (lseek(i, EI_OSABI, SEEK_SET) < 0) {
                    fprintf(stderr, "%s: %s: %s\n", p, argv[optind], strerror(errno));
                    return 1;
                }
                /* write new ABI number */
                if (write(i, &type, 1) < 0) {
                    fprintf(stderr, "%s: %s: %s\n", p, argv[optind], strerror(errno));
                    return 1;
                }
            }
            /* close fd */
            if (close(i) < 0) {
                fprintf(stderr, "%s: %s: %s\n", p, argv[optind], strerror(errno));
                return 1;
            }
            optind++;
        }
    } else {
        usage(p);
        return 1;
    }

    return 0;
}
