/*
 * sishell shellcode processor 0.1
 * Copyright (C) 2004 Shadowinteger <shadowinteger@sentinix.org>
 *
 * This is a re-write of zillion's and EVL's s-proc.c (safemode.org). It has
 * the same syntax (-e -p) but also these additional features:
 *
 *      * -o option to generate octal data (instead of hex)
 *      * -O option to generate ascii-encoded octal data
 *      * -E option to generate detailed example C source code
 *      * scprocessor will attempt to find/guess the offset where the
 *        connect-back ip address and port are located (for reverse shellcode)
 *
 * This processor was specifically written for sishell, it can be used for
 * other shellcodes, but that was not the intention when I wrote this app.
 *
 * The sishell shellcode processor is distributed under the MIT license:
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * $Id: scprocessor.c,v 1.11 2004/03/29 23:40:11 shadow Exp $
 */

#define my_ipaddress    0xfeffff80  /* xored ip 127.0.0.1, as defined in sishell.asm */
#define my_port         0x697a      /* port 31337, network byte order */
#define XOR             0xffffffff  /* sishell xors the ip address, but not the port */

/* max bytes per line */
#define BPERLINE          16
#define BPERLINE_OCTAL    8

/* scprocessor version string */
#define VERSION "0.1"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>

extern char *optarg;
extern int optind, opterr, optopt;

/********************** global variables */
off_t sclen;
void *shellcode;
unsigned char bperline;

/********************** example code */

unsigned char examplecode[] =
    "/* example code below */\n"
    "#include <stdio.h>\n"
    "#include <unistd.h>\n"
    "#include <netinet/in.h>\n"
    "#include <netdb.h>\n"
    "#include <stdint.h>\n"
    "\n"
    "unsigned char ip[] =    /* hardcoded ip */\n"
    "    { 127, 0, 0, 1 };   /* connect to 127.0.0.1 (network byte order) */\n"
    "    /* \"\\x7f\\x00\\x00\\x01\" // if you rather prefer hex */\n"
    "\n"
    "int main(void) {\n"
    "    uint16_t port = 31337;\n"
    "    char hostname[128];\n"
    "    struct hostent *he;\n"
    "\n"
    "    if (gethostname(hostname, sizeof(hostname)) < 0) {\n"
    "        perror(\"gethostname()\");\n"
    "        return 1;\n"
    "    }\n"
    "    if (!(he = gethostbyname(hostname))) {\n"
    "        herror(hostname);\n"
    "        return 1;\n"
    "    }\n"
    "    if (he->h_length != 4) {    /* this is not really needed */\n"
    "        fprintf(stderr, \"only ipv4 addresses please\\n\");\n"
    "        return 1;\n"
    "    }\n"
    "\n"
    "/* choose one of the two conversion methods below:\n"
    " *  // if you want to use a hardcoded ip, do like this:\n"
    " *  *(unsigned long *)&shellcode[IP_OFFSET] = ((uint32_t *)ip)[0] ^ XOR;\n"
    " */\n"
    "    /* if you want to use what gethostbyname() returns, do like this: */\n"
    "    *(unsigned long *)&shellcode[IP_OFFSET] = ((uint32_t *)he->h_addr_list[0])[0] ^ XOR;\n"
    "\n"
    "    /* check that port doesn't contain a null */\n"
    "    if (!(port & 0xff) || !((port & 0xff00)>>8)) {\n"
    "        fprintf(stderr,\"port %u contains null, can't use!\\n\", (uint16_t)port);\n"
    "        return 1;\n"
    "    }\n"
    "\n"
    "    /* the port is simply htons()'ed */\n"
    "    *(uint16_t *)&shellcode[PORT_OFFSET] = htons(port);\n"
    "\n"
    "    /* call shellcode */\n"
    "    ((void (*)())shellcode)();\n"
    "\n"
    "    return 0;\n"
    "}\n";

unsigned char examplecode_octal[] =
    "/* ascii-encoded octal example code below */\n"
    "#include <stdio.h>\n"
    "#include <unistd.h>\n"
    "#include <netdb.h>\n"
    "#include <stdint.h>\n"
    "#include <string.h>\n"
    "\n"
    "unsigned char ip[] =    /* hardcoded ip */\n"
    "    { 127, 0, 0, 1 };   /* connect to 127.0.0.1 (network byte order) */\n"
    "    /* \"\\x7f\\x00\\x00\\x01\" // if you rather prefer hex */\n"
    "\n"
    "int main(void) {\n"
    "    uint16_t port = 31337;\n"
    "    uint32_t xoredip;\n"
    "    char buf[17];\n"
    "    char hostname[128];\n"
    "    struct hostent *he;\n"
    "\n"
    "    if (gethostname(hostname, sizeof(hostname)) < 0) {\n"
    "        perror(\"gethostname()\");\n"
    "        return 1;\n"
    "    }\n"
    "    if (!(he = gethostbyname(hostname))) {\n"
    "        herror(hostname);\n"
    "        return 1;\n"
    "    }\n"
    "    if (he->h_length != 4) {    /* this is not really needed */\n"
    "        fprintf(stderr, \"only ipv4 addresses please\\n\");\n"
    "        return 1;\n"
    "    }\n"
    "\n"
    "/* if your shellcode is ascii-encoded octal, use this method: */\n"
    "    /* if you want to use a hardcoded ip, do like this:\n"
    "    (uint32_t)xoredip = ((uint32_t *)ip)[0] ^ XOR; */\n"
    "\n"
    "    /* if you want to use what gethostbyname() returns, use this: */\n"
    "    (uint32_t)xoredip = ((uint32_t *)he->h_addr_list[0])[0] ^ XOR;\n"
    "\n"
    "    /* use sprintf() to convert ip to octal */\n"
    "    sprintf(buf, \"\\\\%03o\\\\%03o\\\\%03o\\\\%03o\",\n"
    "        ((uint8_t *)&xoredip)[0],\n"
    "        ((uint8_t *)&xoredip)[1],\n"
    "        ((uint8_t *)&xoredip)[2],\n"
    "        ((uint8_t *)&xoredip)[3]\n"
    "        );\n"
    "    memcpy((void *)&shellcode[IP_OFFSET], buf, strlen(buf));\n"
    "\n"
    "    /* make sure port doesn't contain a null */\n"
    "    if (!(port & 0xff) || !((port & 0xff00)>>8)) {\n"
    "        fprintf(stderr,\"port %u contains null, can't use!\\n\", (uint16_t)port);\n"
    "        return 1;\n"
    "    }\n"
    "\n"
    "    /* use sprintf() to convert port to octal */\n"
    "    sprintf(buf, \"\\\\%03o\\\\%03o\",\n"
    "        ((uint8_t *)&port)[1],     /* assume host sys is little-endian */\n"
    "        ((uint8_t *)&port)[0]);\n"
    "    memcpy((void *)&shellcode[PORT_OFFSET], buf, strlen(buf));\n"
    "\n"
    "    /* print ascii octal string of stand alone shellcode */\n"
    "    printf(\"%s\", shellcode);\n"
    "    return 0;\n"
    "}\n";

/********************** functions */

void usage(char *p) {
    printf("usage: %s [-p] [-e] [-o|-O] [-E] [-q] [-c \"comment\"] file file...\n"
           "-p convert inputfile to a C char string\n"
           "-e execute shellcode\n"
           "-o used with -p; produce octal instead of hex\n"
           "-O used with -p; produce ascii-encoded octal\n"
           "-E output detailed example C source code\n"
           "-q don't print header/copyright information in output\n"
           "-c short comment to include after \"char shellcode[] =\"\n", p);
    exit(1);
}

void err(char *msg) {
    perror(msg);
    exit(1);
}

int mallocfile(char *f) {
    FILE *fp;
    struct stat statbuf;

    if (stat(f, &statbuf) < 0)
        err(f);

    sclen = statbuf.st_size;

    if (!(shellcode = malloc((size_t)sclen)))
        err("can't malloc");

    if (!(fp = fopen(f, "r")))
        err(f);

    if (fread(shellcode, 1, sclen, fp) != sclen)
        err(f);

    if (fclose(fp))
        err(f);

    return 0;
}

/********************** main() */

int main(int argc, char **argv) {
    int i, max;
    char octal = 0;
    char dowhat = 0;
    char quiet = 0;
    char example = 0;
    char *comment = NULL;
    uint32_t ip;
    uint16_t port;

    while ((i = getopt(argc, argv, "oOepEqc:h")) != EOF) {
        switch(i) {
            case 'o':
                octal = 1;
                break;
            case 'O':
                octal = 2;
                break;
            case 'p':
                dowhat = 0;
                break;
            case 'e':
                dowhat = 1;
                break;
            case 'E':
                example = 1;
                break;
            case 'q':
                quiet = 1;
                break;
            case 'c':
                comment = (char *)strdup(optarg);
                break;
            case 'h':
            default:
                usage(argv[0]);
                break;
        }
    }

    if (optind < argc) {

        if (!dowhat) {
            if (!quiet)
                printf("/* output by scprocessor " VERSION " (c) 2004 <shadowinteger@sentinix.org> */\n\n");
        }

        while (optind < argc) {
            if (!dowhat) {  /* convert into C string */
                mallocfile(argv[optind]);

                if (!quiet)
                    printf("/* input file: %s */\n", argv[optind]);

                if (octal == 2) {
                    if (!quiet) {
                        printf("/* ascii-encoded octal, for script injection, e.g.:\n"
                               " * printf \"{shellcode}\" > binary\n"
                               " * echo -ne \"{shellcode}\" > binary\n"
                               " * perl -e \"print \\\"{shellcode}\\\"\" > binary\n"
                               " * python -c \"__import__(\\\"sys\\\").stdout.write(\\\"<octal>\\\")\"\n"
                               " */\n");
                    }
                }
                printf("unsigned char shellcode[] =");
                if (comment) {
                    printf(" /* %s */", comment);
                }
                printf("\n    \"");

                max = 0;
                if (octal == 2)
                    bperline = BPERLINE_OCTAL;
                else
                    bperline = BPERLINE;
                for (i = 0; i < sclen; i++) {
                    if (max >= bperline) {
                        printf("\"\n    \"");
                        max = 0;
                    }

                    if (!octal) {
                        printf("\\x%02x", ((unsigned char*)shellcode)[i]);
                    } else if (octal == 1) {    /* print octal */
                        printf("\\%03o", ((unsigned char*)shellcode)[i]);
                    } else if (octal == 2) {    /* print ascii-encoded octal */
                        printf("\\\\%03o", ((unsigned char*)shellcode)[i]);
                    }
                    max++;
                }
                printf("\";\n");

                /* find ip address offset */
                for(i = 0; i <= (sclen-sizeof(uint32_t)); i++) {
                    void *p;
                    p = &((unsigned char *)shellcode)[i];
                    ip = ((uint32_t *)p)[0];
                    if (ip == my_ipaddress) {
                        if (octal == 2) {
                            i = i*4;
                        }
                        printf("#define IP_OFFSET %u    /* _could_ be wrong! */\n", i);

                        break;
                    }
                }
                /* find port offset */
                for(i = 0; i <= (sclen-sizeof(uint16_t)); i++) {
                    void *p;
                    p = &((unsigned char *)shellcode)[i];
                    port = ((uint16_t *)p)[0];
                    if (port == my_port) {
                        if (octal == 2) {
                            i = i*4;
                        }
                        printf("#define PORT_OFFSET %u  /* _could_ be wrong! */\n", i);
                        break;
                    }
                }

                printf("#define XOR 0x%08x   /* number to xor the ip address with */\n\n", XOR);

                /* get next non-option (file) to process */
                optind++;

            } else {    /* else we execute the shellcode */
                mallocfile(argv[optind]); /* mallocfile does exit(1) upon error */
                printf("%s: executing shellcode...\n", argv[0]);
                ((void (*)())shellcode)();
                return 1;   /* should not return here */
            }
        }
    } else {
        usage(argv[0]);
    }

    if (!dowhat) {
        /* output example C code */
        if (example) {
            if (octal == 2)
                printf("%s", examplecode_octal);
            else
                printf("%s", examplecode);
        }
    }

    return 0;
}

/* eof */
